// ------------------------------------------------------------------------------
// DS3231 RTC functions
// ------------------------------------------------------------------------------

#include "ds3231-rtc.h"
#include "z80-mbc2.h"

// ------------------------------------------------------------------------------
// DS3231 RTC variables

extern byte seconds, minutes, hours, dow, day, month, year;
extern byte tempC;    // Temperature (Celsius) encoded in two’s complement integer format

// ------------------------------------------------------------------------------
// RTC Module routines
// ------------------------------------------------------------------------------

byte decToBcd( byte val )
// Convert a binary byte to a two digits BCD byte
{
  return ( ( val / 10 * 16 ) + ( val % 10 ) );
}

// ------------------------------------------------------------------------------

byte bcdToDec( byte val )
// Convert binary coded decimal to normal decimal numbers
{
  return ( ( val / 16 * 10 ) + ( val % 16 ) );
}

// ------------------------------------------------------------------------------

byte rtcInit( void )
{
  byte retVal = 0;

  Wire.beginTransmission( DS3231_RTC );
  if( Wire.endTransmission() == 0 )
  {
    retVal = 1;   // Set to 1 if RTC Module is found
  }

  return retVal;
}

// ------------------------------------------------------------------------------
bool rtcIsRunning( void )
{
  bool retVal = true;
  byte OscStopFlag;

  // Read the "Oscillator Stop Flag"
  Wire.beginTransmission( DS3231_RTC );
  Wire.write( DS3231_STATRG );          // Set the DS3231 Status Register
  Wire.endTransmission();
  Wire.requestFrom( DS3231_RTC, 1 );
  OscStopFlag = Wire.read() & 0x80;    // Read the "Oscillator Stop Flag" 

  if( OscStopFlag )
  {
    // Reset the "Oscillator Stop Flag"
    Wire.beginTransmission( DS3231_RTC );
    Wire.write( DS3231_STATRG );    // Set the DS3231 Status Register
    Wire.write( 0x08 );             // Reset the "Oscillator Stop Flag" (32KHz output left enabled)
    Wire.endTransmission();

    retVal = false;
  } 

  return retVal;
}

// ------------------------------------------------------------------------------

void readRTC( byte *seconds, byte *minutes, byte *hours, byte *dow, byte *day, byte *month, byte *year, byte *tempC )
// Read current date/time binary values and the temprerature (2 complement) from the DS3231 RTC
{
  byte i;

  Wire.beginTransmission( DS3231_RTC );
  Wire.write( DS3231_SECRG );    // Set the DS3231 Seconds Register
  Wire.endTransmission();

  // Read from RTC and convert to binary
  Wire.requestFrom( DS3231_RTC, 18 );
  *seconds = bcdToDec( Wire.read() & 0x7f );
  *minutes = bcdToDec( Wire.read() );
  *hours   = bcdToDec( Wire.read() & 0x3f );
  *dow     = ( Wire.read() - 1 );
  *day   = bcdToDec( Wire.read() );
  *month = bcdToDec( Wire.read() );
  *year  = bcdToDec( Wire.read() );

  for( i = 0; i < 10; i++ )
    Wire.read();    // Jump over 10 registers

  *tempC = Wire.read();
}

// ------------------------------------------------------------------------------

void writeRTC( byte seconds, byte minutes, byte hours, byte day, byte month, byte year )
// Write given date/time binary values to the DS3231 RTC
{
  Wire.beginTransmission( DS3231_RTC );
  Wire.write( DS3231_SECRG );    // Set the DS3231 Seconds Register
  Wire.write( decToBcd( seconds ) );
  Wire.write( decToBcd( minutes ) );
  Wire.write( decToBcd( hours ) );
  Wire.write( dayOfWeek( day, month, (uint16_t)2000 + year ) + 1 );
  Wire.write( decToBcd( day ) );
  Wire.write( decToBcd( month ) );
  Wire.write( decToBcd( year ) );
  Wire.endTransmission();
}

// ------------------------------------------------------------------------------

void printDateTime( byte readSourceFlag )
// Print to serial the current date/time from the global variables.
//
// Flag readSourceFlag [0..1] usage:
//    If readSourceFlag = 0 the RTC read is not done
//    If readSourceFlag = 1 the RTC read is done (global variables are updated)
{
  char dowStr[7][4] = { "Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun" };

  if( readSourceFlag )
  {
    readRTC( &seconds, &minutes, &hours, &dow, &day, &month, &year, &tempC );
  }  
  
  print2digit( day );
  Serial.print( "/" );
  print2digit( month );
  Serial.print( "/" );
  print2digit( year );
  Serial.print( " " );
  if( dow <= 6 ) { Serial.print( dowStr[dow] ); }
  else { Serial.print( dow ); }
  Serial.print( " " );
  print2digit( hours );
  Serial.print( ":" );
  print2digit( minutes );
  Serial.print( ":" );
  print2digit( seconds );
}

// ------------------------------------------------------------------------------

void print2digit( byte data )
// Print a byte [0..99] using 2 digit with leading zeros if needed
{
  if( data < 10 ) { Serial.print( "0" ); }
  Serial.print( data );
}

// ------------------------------------------------------------------------------

bool isLeapYear( uint16_t year )
// Check if the year is a leap year.
// Returns true if it is leap, false otherwise.
{
  if( ( ( year % 4 ) == 0 ) && ( ( year % 100 ) != 0 ) || ( ( year % 400 ) == 0 ) ) 
    return true;
  else
    return false;
}

// ------------------------------------------------------------------------------

uint8_t dayOfWeek( uint8_t day, uint8_t month, uint16_t year )
{
   // Zeller's congruence for the Gregorian calendar. 
   // With 0=Monday, ... 5=Saturday, 6=Sunday         

   if( month < 3 )
   {
      month += 12;
      year--;
   }

  return ((13*month+3)/5 + day + year + year/4 - year/100 + year/400) % 7;
}

// ------------------------------------------------------------------------------

uint8_t daysOfMonth( uint8_t month, uint16_t year )
{
  uint8_t days = 30;

  // 1, 3, 5, 7
  if( ( ( month < 8 ) && ( (month % 2 ) != 0 ) ) || ( ( month >= 8 ) && ( ( month % 2 ) == 0 ) ) ) { days = 31; }

  if( month == 2 )
  {
    if( isLeapYear( year ) == true ) { days = 29; }
    else { days = 28; }
  }

  return days;
}

// ------------------------------------------------------------------------------

void ChangeRTC( void )
// Change manually the RTC Date/Time from keyboard
{
  unsigned long timeStamp;  // Timestamp for led blinking
  char inChar;              // Input char from serial
  byte step;                // Temporary variable (buffer)

  // Read RTC
  readRTC( &seconds, &minutes, &hours, &dow, &day, &month, &year, &tempC );

  // Change RTC date/time from keyboard
  step = 0;
  Serial.println( "IOS: RTC manual setting:" );
  Serial.println( "Press T/U to increment +10/+1 or CR to accept" );
  do
  {
    do
    {
      Serial.print( " " );
      switch( step )
      {
        case 0:
          Serial.print( "Year -> " );
          print2digit( year );
          break;

        case 1:
          Serial.print( "Month -> " );
          print2digit( month );
          break;

        case 2:
          Serial.print( "             " );
          Serial.write( 13 );
          Serial.print( " Day -> " );
          print2digit( day );
          break;

        case 3:
          Serial.print( "Hours -> " );
          print2digit( hours );
          break;

        case 4:
          Serial.print( "Minutes -> " );
          print2digit( minutes );
          break;

        case 5:
          Serial.print( "Seconds -> " );
          print2digit( seconds );
          break;
      }

      timeStamp = millis();
      do
      {
        blinkIOSled( &timeStamp );
        inChar = Serial.read();
      } while( ( inChar != 'u' ) && ( inChar != 'U' ) && ( inChar != 't' ) && ( inChar != 'T' ) && ( inChar != 13 ) );

      if( ( inChar == 'u' ) || ( inChar == 'U' ) )
      {
        // Change units
        switch( step )
        {
          case 0:
            year++;
            if( year > 99 ) { year = 0; }
            break;

          case 1:
            month++;
            if( month > 12 ) { month = 1; }
            break;

          case 2:
            day++;
            if( day > daysOfMonth( month, year ) ) { day = 1; }
            break;

          case 3:
            hours++;
            if( hours > 23 ) { hours = 0; }
            break;

          case 4:
            minutes++;
            if( minutes > 59 ) { minutes = 0; }
            break;

          case 5:
            seconds++;
            if( seconds > 59 ) { seconds = 0; }
            break;
        }
      }

      if( ( inChar == 't' ) || ( inChar == 'T' ) )
      {
        // Change tens
        switch( step )
        {
          case 0:
            year = year + 10;
            if( year > 99 ) { year = year - ( year / 10 ) * 10; }
            break;

          case 1:
            if( month > 10 ) { month = month - 10; }
            else if( month < 3 ) { month = month + 10; }
            break;

          case 2:
            day = day + 10;
            if( day > daysOfMonth( month, year ) ) { day = day - ( day / 10 ) * 10; }
            if( day == 0 ) { day = 1; }
            break;

          case 3:
            hours = hours + 10;
            if( hours > 23 ) { hours = hours - ( hours / 10 ) * 10; }
            break;

          case 4:
            minutes = minutes + 10;
            if( minutes > 59 ) { minutes = minutes - ( minutes / 10 ) * 10; }
            break;

          case 5:
            seconds = seconds + 10;
            if( seconds > 59 ) { seconds = seconds - ( seconds / 10 ) * 10; }
            break;
        }
      }

      Serial.write( 13 );
    } while( inChar != 13 );

    step++;
  } while( step < 6 );

  // Write new date/time into the RTC
  writeRTC( seconds, minutes, hours, day, month, year );
  Serial.println( F( " ...done      " ) );
  Serial.print( F( "IOS: RTC date/time updated (" ) );
  printDateTime( 1 );
  Serial.println( ")" );
}
