// SD Disk routines (FAT16 and FAT32 filesystems supported) using the PetitFS library.
// For more info about PetitFS see here: http://elm-chan.org/fsw/ff/00index_p.html

#include "sd-card.h"

// ------------------------------------------------------------------------------

byte mountSD( FATFS *fatFs )
// Mount a volume on SD:
// *  "fatFs" is a pointer to a FATFS object (PetitFS library)
// The returned value is the resulting status (0 = ok, otherwise see printErrSD())
{
    return pf_mount( fatFs );
}

// ------------------------------------------------------------------------------

byte openSD( const char *fileName )
// Open an existing file on SD:
// *  "fileName" is the pointer to the string holding the file name (8.3 format)
// The returned value is the resulting status (0 = ok, otherwise see printErrSD())
{
    return pf_open( fileName );
}

// ------------------------------------------------------------------------------

byte readSD( void *buffSD, byte *numReadBytes )
// Read one "segment" (32 bytes) starting from the current sector (512 bytes) of the opened file on SD:
// *  "BuffSD" is the pointer to the segment buffer;
// *  "numReadBytes" is the pointer to the variables that store the number of read bytes;
//     if < 32 (including = 0) an EOF was reached).
// The returned value is the resulting status (0 = ok, otherwise see printErrSD())
//
// NOTE1: Each SD sector (512 bytes) is divided into 16 segments (32 bytes each); to read a sector you need to
//        to call readSD() 16 times consecutively
//
// NOTE2: Past current sector boundary, the next sector will be pointed. So to read a whole file it is sufficient
//        call readSD() consecutively until EOF is reached
{
    UINT numBytes;
    byte errcode;
    errcode       = pf_read( buffSD, 32, &numBytes );
    *numReadBytes = (byte)numBytes;
    return errcode;
}

// ------------------------------------------------------------------------------

byte writeSD( void *buffSD, byte *numWrittenBytes )
// Write one "segment" (32 bytes) starting from the current sector (512 bytes) of the opened file on SD:
// *  "BuffSD" is the pointer to the segment buffer;
// *  "numWrittenBytes" is the pointer to the variables that store the number of written bytes;
//     if < 32 (including = 0) an EOF was reached.
// The returned value is the resulting status (0 = ok, otherwise see printErrSD())
//
// NOTE1: Each SD sector (512 bytes) is divided into 16 segments (32 bytes each); to write a sector you need to
//        to call writeSD() 16 times consecutively
//
// NOTE2: Past current sector boundary, the next sector will be pointed. So to write a whole file it is sufficient
//        call writeSD() consecutively until EOF is reached
//
// NOTE3: To finalize the current write operation a writeSD(NULL, &numWrittenBytes) must be called as last action
{
    UINT numBytes;
    byte errcode;
    if( buffSD != NULL )
    {
        errcode = pf_write( buffSD, 32, &numBytes );
    }
    else
    {
        errcode = pf_write( 0, 0, &numBytes );
    }
    *numWrittenBytes = (byte)numBytes;
    return errcode;
}

// ------------------------------------------------------------------------------

byte seekSD( word sectNum )
// Set the pointer of the current sector for the current opened file on SD:
// *  "sectNum" is the sector number to set. First sector is 0.
// The returned value is the resulting status (0 = ok, otherwise see printErrSD())
//
// NOTE: "secNum" is in the range [0..16383], and the sector addressing is continuos inside a "disk file";
//       16383 = (512 * 32) - 1, where 512 is the number of emulated tracks, 32 is the number of emulated sectors
//
{
    byte i;
    return pf_lseek( ( (unsigned long)sectNum ) << 9 );
}

// ------------------------------------------------------------------------------

void printErrSD( byte opType, byte errCode, const char *fileName )
// Print the error occurred during a SD I/O operation:
//  * "OpType" is the operation that generated the error (0 = mount, 1= open, 2 = read,
//     3 = write, 4 = seek);
//  * "errCode" is the error code from the PetitFS library (0 = no error);
//  * "fileName" is the pointer to the file name or NULL (no file name)
//
// ........................................................................
//
// Errors legend (from PetitFS library) for the implemented operations:
//
// ------------------
// mountSD():
// ------------------
// NOT_READY
//     The storage device could not be initialized due to a hard error or no medium.
// DISK_ERR
//     An error occured in the disk read function.
// NO_FILESYSTEM
//     There is no valid FAT partition on the drive.
//
// ------------------
// openSD():
// ------------------
// NO_FILE
//     Could not find the file.
// DISK_ERR
//     The function failed due to a hard error in the disk function, a wrong FAT structure or an internal error.
// NOT_ENABLED
//     The volume has not been mounted.
//
// ------------------
// readSD() and writeSD():
// ------------------
// DISK_ERR
//     The function failed due to a hard error in the disk function, a wrong FAT structure or an internal error.
// NOT_OPENED
//     The file has not been opened.
// NOT_ENABLED
//     The volume has not been mounted.
//
// ------------------
// seekSD():
// ------------------
// DISK_ERR
//     The function failed due to an error in the disk function, a wrong FAT structure or an internal error.
// NOT_OPENED
//     The file has not been opened.
//
// ........................................................................
{
    if( errCode )
    {
        Serial.print( "\r\nIOS: SD error " );
        Serial.print( errCode );
        Serial.print( " (" );
        switch( errCode )
        // See PetitFS implementation for the codes
        {
            case 1:
                Serial.print( "DISK_ERR" );
                break;
            case 2:
                Serial.print( "NOT_READY" );
                break;
            case 3:
                Serial.print( "NO_FILE" );
                break;
            case 4:
                Serial.print( "NOT_OPENED" );
                break;
            case 5:
                Serial.print( "NOT_ENABLED" );
                break;
            case 6:
                Serial.print( "NO_FILESYSTEM" );
                break;
            default:
                Serial.print( "UNKNOWN" );
        }
        Serial.print( " on " );
        switch( opType )
        {
            case 0:
                Serial.print( "MOUNT" );
                break;
            case 1:
                Serial.print( "OPEN" );
                break;
            case 2:
                Serial.print( "READ" );
                break;
            case 3:
                Serial.print( "WRITE" );
                break;
            case 4:
                Serial.print( "SEEK" );
                break;
            default:
                Serial.print( "UNKNOWN" );
        }
        Serial.print( " operation" );
        if( fileName )
        // Not a NULL pointer, so print file name too
        {
            Serial.print( " - File: " );
            Serial.print( fileName );
        }
        Serial.println( ")" );
    }
}
