// ------------------------------------------------------------------------------
// Z80-MB2 hardware functions
// ------------------------------------------------------------------------------

#include "z80-mbc2.h"

// ------------------------------------------------------------------------------
extern unsigned long timeStamp;    // Timestamp for led blinking

// ------------------------------------------------------------------------------

byte mbc2_init( void )
{
  byte retVal = 0;

  // Initialize RESET_ and WAIT_RES_
  pinMode( RESET_, OUTPUT );        // Configure RESET_ and set it ACTIVE
  digitalWrite( RESET_, LOW );
  pinMode( WAIT_RES_, OUTPUT );     // Configure WAIT_RES_ and set it ACTIVE to reset the WAIT FF (U1C/D)
  digitalWrite( WAIT_RES_, LOW );

  // Check USER Key for boot mode changes
  pinMode( USER, INPUT_PULLUP );    // Read USER Key to enter into the boot mode selection
  if( !digitalRead( USER ) ) { retVal = 1; }

  // Initialize USER,  INT_, RAM_CE2, and BUSREQ_
  pinMode( USER, OUTPUT );          // USER led OFF
  digitalWrite( USER, HIGH );
  pinMode( INT_, INPUT_PULLUP );    // Configure INT_ and set it NOT ACTIVE
  pinMode( INT_, OUTPUT );
  digitalWrite( INT_, HIGH );
  pinMode( RAM_CE2, OUTPUT );       // Configure RAM_CE2 as output
  digitalWrite( RAM_CE2, HIGH );    // Set RAM_CE2 active
  pinMode( WAIT_, INPUT );          // Configure WAIT_ as input
  pinMode( BUSREQ_, INPUT_PULLUP ); // Set BUSREQ_ HIGH
  pinMode( BUSREQ_, OUTPUT );
  digitalWrite( BUSREQ_, HIGH );

  // Initialize D0-D7, AD0, MREQ_, RD_ and WR_
  DDRA  = 0x00;    // Configure Z80 data bus D0-D7 (PA0-PA7) as input with pull-up
  PORTA = 0xFF;
  pinMode( MREQ_, INPUT_PULLUP );   // Configure MREQ_ as input with pull-up
  pinMode( RD_, INPUT_PULLUP );     // Configure RD_ as input with pull-up
  pinMode( WR_, INPUT_PULLUP );     // Configure WR_ as input with pull-up
  pinMode( AD0, INPUT_PULLUP );

  // Initialize the Logical RAM Bank (32KB) to map into the lower half of the Z80 addressing space
  pinMode( BANK0, OUTPUT );         // Set RAM Logical Bank 1 (Os Bank 0)
  digitalWrite( BANK0, HIGH );
  pinMode( BANK1, OUTPUT );
  digitalWrite( BANK1, LOW );

  // Initialize CLK (single clock pulses mode) and reset the Z80 CPU
  pinMode( CLK, OUTPUT );    // Set CLK as output

  singlePulsesResetZ80();    // Reset the Z80 CPU using single clock pulses

  // Initialize MCU_RTS and MCU_CTS and reset uTerm (A071218-R250119) if present
  pinMode( MCU_CTS_, INPUT_PULLUP );  // Parked (not used)
  pinMode( MCU_RTS_, OUTPUT );
  digitalWrite( MCU_RTS_, LOW );      // Reset uTerm (A071218-R250119)
  delay( 100 );
  digitalWrite( MCU_RTS_, HIGH );
  delay( 500 ); 

  return retVal; 
}

// ------------------------------------------------------------------------------

void blinkIOSled( unsigned long *timestamp )
// Blink led IOS using a timestamp
{
  if( ( millis() - *timestamp ) > 200 )
  {
    digitalWrite( LED_IOS, !digitalRead( LED_IOS ) );
    *timestamp = millis();
  }
}

// ------------------------------------------------------------------------------

bool isJtagActive( void )
{
  uint8_t mcuReg;
  mcuReg = MCUCSR;

  return (mcuReg & 0x80);
}

// ------------------------------------------------------------------------------

// ------------------------------------------------------------------------------
// Z80 bootstrap routines
// ------------------------------------------------------------------------------

void pulseClock( byte numPulse )
// Generate <numPulse> clock pulses on the Z80 clock pin.
// The steady clock level is LOW, e.g. one clock pulse is a 0-1-0 transition
{
  byte i;
  for( i = 0; i < numPulse; i++ )
  // Generate one clock pulse
  {
    // Send one impulse (0-1-0) on the CLK output
    digitalWrite( CLK, HIGH );
    digitalWrite( CLK, LOW );
  }
}

// ------------------------------------------------------------------------------

void loadByteToRAM( byte value )
// Load a given byte to RAM using a sequence of two Z80 instructions forced on the data bus.
// The RAM_CE2 signal is used to force the RAM in HiZ, so the Atmega can write the needed instruction/data
//  on the data bus. Controlling the clock signal and knowing exactly how many clocks pulse are required it
//  is possible control the whole loading process.
// In the following "T" are the T-cycles of the Z80 (See the Z80 datashet).
// The two instruction are "LD (HL), n" and "INC (HL)".
{

  // Execute the LD(HL),n instruction (T = 4+3+3). See the Z80 datasheet and manual.
  // After the execution of this instruction the <value> byte is loaded in the memory address pointed by HL.
  pulseClock( 1 );                  // Execute the T1 cycle of M1 (Opcode Fetch machine cycle)
  digitalWrite( RAM_CE2, LOW );     // Force the RAM in HiZ (CE2 = LOW)
  DDRA  = 0xFF;                     // Configure Z80 data bus D0-D7 (PA0-PA7) as output
  PORTA = LD_HL;                    // Write "LD (HL), n" opcode on data bus
  pulseClock( 2 );                  // Execute T2 and T3 cycles of M1
  DDRA  = 0x00;                     // Configure Z80 data bus D0-D7 (PA0-PA7) as input...
  PORTA = 0xFF;                     // ...with pull-up
  pulseClock( 2 );                  // Complete the execution of M1 and execute the T1 cycle of the following
                                    // Memory Read machine cycle
  DDRA  = 0xFF;                     // Configure Z80 data bus D0-D7 (PA0-PA7) as output
  PORTA = value;                    // Write the byte to load in RAM on data bus
  pulseClock( 2 );                  // Execute the T2 and T3 cycles of the Memory Read machine cycle
  DDRA  = 0x00;                     // Configure Z80 data bus D0-D7 (PA0-PA7) as input...
  PORTA = 0xFF;                     // ...with pull-up
  digitalWrite( RAM_CE2, HIGH );    // Enable the RAM again (CE2 = HIGH)
  pulseClock( 3 );                  // Execute all the following Memory Write machine cycle

  // Execute the INC(HL) instruction (T = 6). See the Z80 datasheet and manual.
  // After the execution of this instruction HL points to the next memory address.
  pulseClock( 1 );                  // Execute the T1 cycle of M1 (Opcode Fetch machine cycle)
  digitalWrite( RAM_CE2, LOW );     // Force the RAM in HiZ (CE2 = LOW)
  DDRA  = 0xFF;                     // Configure Z80 data bus D0-D7 (PA0-PA7) as output
  PORTA = INC_HL;                   // Write "INC(HL)" opcode on data bus
  pulseClock( 2 );                  // Execute T2 and T3 cycles of M1
  DDRA  = 0x00;                     // Configure Z80 data bus D0-D7 (PA0-PA7) as input...
  PORTA = 0xFF;                     // ...with pull-up
  digitalWrite( RAM_CE2, HIGH );    // Enable the RAM again (CE2 = HIGH)
  pulseClock( 3 );                  // Execute all the remaining T cycles
}

// ------------------------------------------------------------------------------

void loadHL( word value )
// Load "value" word into the HL registers inside the Z80 CPU, using the "LD HL,nn" instruction.
// In the following "T" are the T-cycles of the Z80 (See the Z80 datashet).
{
  // Execute the LD dd,nn instruction (T = 4+3+3), with dd = HL and nn = value. See the Z80 datasheet and manual.
  // After the execution of this instruction the word "value" (16bit) is loaded into HL.
  pulseClock( 1 );                  // Execute the T1 cycle of M1 (Opcode Fetch machine cycle)
  digitalWrite( RAM_CE2, LOW );     // Force the RAM in HiZ (CE2 = LOW)
  DDRA  = 0xFF;                     // Configure Z80 data bus D0-D7 (PA0-PA7) as output
  PORTA = LD_HLnn;                  // Write "LD HL, n" opcode on data bus
  pulseClock( 2 );                  // Execute T2 and T3 cycles of M1
  DDRA  = 0x00;                     // Configure Z80 data bus D0-D7 (PA0-PA7) as input...
  PORTA = 0xFF;                     // ...with pull-up
  pulseClock( 2 );                  // Complete the execution of M1 and execute the T1 cycle of the following
                                    // Memory Read machine cycle
  DDRA  = 0xFF;                     // Configure Z80 data bus D0-D7 (PA0-PA7) as output
  PORTA = lowByte( value );         // Write first byte of "value" to load in HL
  pulseClock( 3 );                  // Execute the T2 and T3 cycles of the first Memory Read machine cycle
                                    // and T1, of the second Memory Read machine cycle
  PORTA = highByte( value );        // Write second byte of "value" to load in HL
  pulseClock( 2 );                  // Execute the T2 and T3 cycles of the second Memory Read machine cycle
  DDRA  = 0x00;                     // Configure Z80 data bus D0-D7 (PA0-PA7) as input...
  PORTA = 0xFF;                     // ...with pull-up
  digitalWrite( RAM_CE2, HIGH );    // Enable the RAM again (CE2 = HIGH)
}

// ------------------------------------------------------------------------------

void singlePulsesResetZ80( void )
// Reset the Z80 CPU using single pulses clock
{
  digitalWrite( RESET_, LOW );     // Set RESET_ active
  pulseClock( 6 );                 // Generate twice the needed clock pulses to reset the Z80
  digitalWrite( RESET_, HIGH );    // Set RESET_ not active
  pulseClock( 2 );                 // Needed two more clock pulses after RESET_ goes HIGH
}

